/* eslint-disable no-unused-vars, import/extensions */
import {
    Gantt, Panel, ProjectModel, Toast, Timeline
    //, StringHelper, VersionGrid, Container, MessageDialog, ChangeLogTransactionModel, VersionModel
    , DateHelper
    //, GlobalEvents, WidgetHelper, DomHelper
} from '../build/gantt.module.js';

import GlobalEvents from '../lib/Core/GlobalEvents.js';
import WidgetHelper from '../lib/Core/helper/WidgetHelper.js';
import DomHelper from '../lib/Core/helper/DomHelper.js';

import Task from './lib/models/Task.js';

import Dependency from './lib/models/Dependency.js';
import Calendar from './lib/models/Calendar.js';
import Resource from './lib/models/Resource.js';
import Assignment from './lib/models/Assignment.js';
import GanttToolbar from './lib/GanttToolbar.js';
import BackendTools from './lib/BackendTools.js';

const project = window.project = new ProjectModel({
    taskModelClass: Task,
    dependencyModelClass: Dependency,
    resourceModelClass: Resource,
    assignmentModelClass: Assignment,
    calendarModelClass: Calendar,
    transport: {
        load: {
            url: '../_datasets/launch-saas.json'
        }
    },
    autoLoad: true,
    taskStore: {
        wbsMode: 'auto'
    },
    // The State TrackingManager which the UndoRedo widget in the toolbar uses
    stm: {
        autoRecord: true
    },
    // Reset Undo / Redo after each load
    resetUndoRedoQueuesAfterLoad: true
});


const
    DATE_FORMAT = 'MMM DD, YYYY',
    serializeDate = (value) => DateHelper.format(value, DATE_FORMAT);

//Start Export PDF

const headerTpl = ({ currentPage, totalPages }) => `
    <div class="demo-export-header">
        <img src="resources/logo.png"/>
        <dl>
            <dt>Date: ${DateHelper.format(new Date(), DATE_FORMAT)}</dt>
            <dd>${totalPages ? `Page: ${currentPage + 1}/${totalPages}` : ''}</dd>
        </dl>
    </div>`;

const footerTpl = () => '';//<div class="demo-export-footer"><h3> 2020 Bryntum AB</h3></div>


//End Export PDF

const DAY_DATE_FORMAT = "ddd MMM DD, YYYY";


const gantt = window.gantt = new Gantt({
    project,
    //flex: 2,
    dependencyIdField: 'sequenceNumber',


    weekStartDay: 1,
    columns: [
        { type: 'sequence', width: 20 },
        { type: 'wbs' },
        { type: 'code', width: 90 },
        { type: 'percentdone', showCircle: true, text: '%', width: 70 },
        { type: 'name', width: 170 },
        { text: 'Milestone', type: 'milestone' },
        { text: 'Show In Timeline', type: 'showintimeline' },
        { type: 'startdate', format: DAY_DATE_FORMAT },
        { type: 'enddate', format: DAY_DATE_FORMAT },
        { type: 'duration' },
        { type: 'calendar' },
        /*{ type: 'totalslack'},*/
        { type: 'effort' },
        { type: 'resourceassignment', width: 120, showAvatars: true },
        {
            type: 'predecessor',
            width: 112
        },
        {
            type: 'successor',
            width: 112
        },
        { type: 'schedulingmodecolumn' },
        { type: 'manuallyscheduled' },

        { type: 'constrainttype' },
        { type: 'constraintdate', format: DAY_DATE_FORMAT },
        { type: 'deadlinedate', format: DAY_DATE_FORMAT },

        { type: 'earlystartdate', format: DAY_DATE_FORMAT },
        { type: 'earlyenddate', format: DAY_DATE_FORMAT },
        { type: 'addnew' }
    ],

    subGridConfigs: {
        locked: {
            flex: 4
        },
        normal: {
            flex: 3
        }
    },

    loadMask: 'Loading tasks...',

    columnLines: false,
    selectionMode: {
        cell: true,
        column: true,
        dragSelect: true,
        rowNumber: true
    },
    features: {
        indicators: {
            items: {
                deadline: true,
                earlyDates: false,
                lateDates: false,
                constraintDate: false
            }
        },
        //sort:"OrderId",
        rollups: {
            disabled: true
        },
        progressLine: {
            disabled: true,
        },

        cellEdit: {
            autoEdit: true, 
            multiEdit: true 
        },
        fillHandle: true,
        cellCopyPaste: true,
        taskCopyPaste: {
            allowNativeClipboard: true
        },

        taskMenu: {
            items: {
                convertToMilestone: true,

            },
            processItems({ taskRecord, items }) {

                if (taskRecord.isMilestone) {
                    items.convertToMilestone = false;
                }
            },

        },
        filter: true,
        dependencyEdit: true,
        timeRanges: {
            showCurrentTimeLine: true
        },

        labels: {
            left: {
                field: 'name',
                editor: {
                    type: 'textfield'
                }
            }
        }
    },

    tbar: {
        items: [
            {
                type: 'widget',
                cls: 'label',
                html: 'Theme:'
            },
            {
                type: 'buttonGroup',
                toggleGroup: true,
                items: ['Stockholm', 'Material', 'Classic-Light', 'Classic', 'Classic-Dark'].map(name => {
                    return {
                        id: name.toLowerCase(),
                        text: name,
                        pressed: DomHelper.themeInfo.name === name
                    };
                }),
                onAction({ source: button }) {
                    DomHelper.setTheme(button.text);
                }
            },
        ]
    }
});



// Add Save / Load / Reset buttons toolbar and server data load/sync handlers
const backendTools = new BackendTools(gantt);


new Panel({
    appendTo: 'bryntum-task-container',
    layout: 'fit',
    items: [
        gantt
    ],
    tbar: new GanttToolbar({ gantt })
});

GlobalEvents.on({
    theme(themeChangeEvent) {
        WidgetHelper.getById(themeChangeEvent.theme).pressed = true;
    }
});

// console.time("load data");
project.load().then(() => {
    // console.timeEnd("load data");
    const stm = gantt.project.stm;

    stm.enable();
    stm.autoRecord = true;

}).catch(({ response, responseText }) => {

    if (response && response.message) {
        Toast.show({
            html: `${response.message}<br>
                    <b>Please make sure that you've read readme.md file carefully
                    and setup the database connection accordingly.</b>`,
            color: 'b-red',
            style: 'color:white',
            timeout: 0
        });
    }
});

//Start Timeline
const setTimelineHeight = ({ source: button }) => {
    timeline.element.style.height = '';

    ['large', 'medium', 'small'].forEach((cls) => timeline.element.classList.remove(cls));

    timeline.element.classList.add(button.text.toLowerCase());
};


const timeline = window.timeline = new Timeline({
    appendTo: 'bryntum-task-timeline',

    height: '', // Override default to let it be controlled by CSS

    eventStyle: 'plain',
    eventColor: 'gantt-green',

    project,

    dependencyIdField: 'sequenceNumber',

    tbar: [
        'Timeline size',
        {
            type: 'buttonGroup',
            toggleGroup: true,
            items: [
                {
                    text: 'Small'
                }, {
                    text: 'Medium',
                    pressed: true
                }, {
                    text: 'Large'
                }
            ],
            onAction: setTimelineHeight
        }
    ]
});

//End Timeine


